/*

This file is part of jHaushalt.

jHaushalt is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

jHaushalt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with jHaushalt; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

(C)opyright 2002-2005 Dr. Lars H. Hahn

*/

package haushalt.daten;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.ArrayList;

/**
 * Die Register dienen als Container fr die Buchung. Beispiele
 * fr Register sind "Girokonto" oder "Bargeld".
 * @author Dr. Lars H. Hahn
 * @version 2.0/2004.08.22
 */
class Register {
	private final boolean DEBUG = false;

	private final String name;
	private final ArrayList buchungen;

	Register(String name) {
		this.name = name;
    buchungen = new ArrayList();
  }

	public String toString() {
		return name;
	}

	/**
	 * Liefert die Anzahl der im Register gespeicherten Buchungen.
	 * @return Anzahl der Buchungen
	 */
	int getAnzahlBuchungen() {
		return buchungen.size();
	}
	
	/**
	 * Liefert die Buchung an der angegebenen Position.
	 * @param nr Position der gesuchten Buchung
	 * @return gesuchte Buchung
	 */
	AbstractBuchung getBuchung(int nr) {
	  if(nr == getAnzahlBuchungen())
	    return new StandardBuchung();
		return (AbstractBuchung)buchungen.get(nr);
	}
	
	/**
	 * Wird bentigt um alte Buchungen zu lschen.
	 * @param nr Position bis zu der Buchungen gelscht werden
	 */
	void removeBisBuchung(int nr) {
		for(int i=0;i<=nr;i++)
			buchungen.remove(0);
		// mit jedem remove verndert sich der Index -> immer erstes Element lschen
	}

	/**
	 * Lscht eine Buchung.
	 * @param nr
	 */
	void entferneBuchung(int nr) {
		if(getBuchung(nr).getClass() == Umbuchung.class) {
			// Die Buchung muss auch aus dem Partner-Register entfernt werden.
			Register[] register = (Register[]) getBuchung(nr).getKategorie();
			for(int i=0; i<2; i++)
				if(register[i] != this)
					register[i].buchungen.remove(getBuchung(nr));
		}
	  if(nr < getAnzahlBuchungen())
	    buchungen.remove(nr);
	}
	
	/**
	 * Entfernt eine Umbuchung. Hiermit kann eine Umbuchung bei nderung seines
	 * Quell- oder Zielregisters sich aus dem alten Register entfernen.
	 * @param buchung
	 */
	void loescheUmbuchung(Umbuchung buchung) {
		buchungen.remove(buchung);
	}
  
  /**
   * Sortiert die bergebene Buchung in das Register ein.
   * Es wird aufsteigend nach Buchungsdatum sortiert.
   * @param buchung einzusortierende Buchung
   * @return Einfge-Position 
   */
  int einsortierenBuchung(AbstractBuchung buchung) {
    int size = buchungen.size();
    int pos = -1;
    for(int i=0; i<size; i++)
      if(buchung.compareTo(buchungen.get(i)) >= 0)
        pos = i;
    if(pos == size-1) {  // ans Ende
      buchungen.add(buchung);
    }
    else {              // neue Buchung einfuegen
      buchungen.add(pos+1, buchung);
    }
		if(DEBUG)
			System.out.println("Register "+name+": Buchung "+buchung.getDatum()+"/"+buchung.getText()+"/"+buchung.getWert()+" an Positon "+(pos+1)+" einsortiert.");
		return pos+1;
  }
  
  /**
   * Sortiert die angegebene Buchung neu ein.
   * Dies wird notwendig, wenn das Datum der Buchung gendert wurde.
   * @param nr Index der genderten Buchung
   * @return Einfge-Position 
   */
  int buchungNeusortieren(int nr) {
    AbstractBuchung buchung = (AbstractBuchung) buchungen.remove(nr);
    return einsortierenBuchung(buchung);
  }

	/**
	 * Fgt die Buchungen aus einem anderen Register diesem Register hinzu und
	 * lscht sie dann.
	 * @param registerZumLoeschen Register aus dem die Buchungen bernommen werden
	 */
	void registerVereinigen(Register registerZumLoeschen) {
			while(registerZumLoeschen.getAnzahlBuchungen() > 0) {
				if(registerZumLoeschen.buchungen.get(0).getClass() == Umbuchung.class) {
					Umbuchung umbuchung = (Umbuchung)registerZumLoeschen.buchungen.get(0);
					Register[] alteKategorie = (Register[]) umbuchung.getKategorie();
					Register[] neueKategorie = new Register[2];
					
					if(alteKategorie[0] == alteKategorie[1]) {
					  // Umbuchung: alte Selbstbuchung
					  neueKategorie[0] = this;
					  neueKategorie[1] = this;
						umbuchung.setKategorie(neueKategorie);
						// -> automatisch lschen und neu einfgen
					}
					else {
					  // normale Umbuchung
						for(int i=0; i<2; i++)
						  if(alteKategorie[i] == registerZumLoeschen)
						    // altes durch neues Register ersetzten 
						    neueKategorie[i] = this;
						  else
						    neueKategorie[i] = alteKategorie[i];
						if(neueKategorie[0] != neueKategorie[1])
							umbuchung.setKategorie(neueKategorie);
							// nur automatisch (!) einsortieren, wenn keine Selbstbuchung entstanden ist
						else {
						  // sonst loeschen:
						  neueKategorie[0].loescheUmbuchung(umbuchung);
							registerZumLoeschen.buchungen.remove(0);
						}
					}
				}
				else { 
				  // StandardBuchung + SplitBuchung
					einsortierenBuchung((AbstractBuchung)registerZumLoeschen.buchungen.get(0));
					registerZumLoeschen.buchungen.remove(0);
				}
			}
		}

		/**
		 * Liefert den Saldo des Registers bis zum angegebenen Datum (exklusiv).
		 * @param datum
		 * @return
		 */
		Euro getSaldo(Datum datum) {
			Euro saldo = new Euro();
			int i = 0;
			while((i < buchungen.size()) &&
				(datum.compareTo(getBuchung(i).getDatum()) > 0)) {
				if(getBuchung(i).getClass() == Umbuchung.class) {
					Register[] register = (Register[]) getBuchung(i).getKategorie();
					if((register[0] == this) && (register[0] != register[1]))
  					saldo = saldo.sub(getBuchung(i).getWert());
	  				// Falls dieses Register Quelle einer Umbuchung ist: Subtrahieren!
					// Selbstbuchugen aber nicht!
	  			else
	  			  saldo.sum(getBuchung(i).getWert());
				}
				else { // StandardBuchung + SplitBuchung
					saldo.sum(getBuchung(i).getWert());
				}
				i++;
			}
			return saldo;
		}

		/**
		 * Liefert den Saldo des Registers bis zur angegebenen Zeile.
		 * @param row
		 * @return
		 */
		Euro getSaldo(int row) {
			Euro saldo = new Euro();
			for(int i=0; i<=row; i++)
				if(getBuchung(i).getClass() == Umbuchung.class) {
					Register[] register = (Register[]) getBuchung(i).getKategorie();
					if((register[0] == this) && (register[0] != register[1]))
						saldo = saldo.sub(getBuchung(i).getWert());
						// Falls dieses Register Quelle einer Umbuchung ist: Subtrahieren!
						// Selbstbuchugen aber nicht!
					else
					  saldo.sum(getBuchung(i).getWert());
				}
				else { // StandardBuchung + SplitBuchung
					saldo.sum(getBuchung(i).getWert());
				}
			return saldo;
		}

    
    // -- E/A-Funktionen -------------------------------------------------------
    
		void laden(DataInputStream in, Datenbasis db)
			throws IOException {
			int size = in.readInt();
			buchungen.ensureCapacity(size);
			for(int i=0; i<size; i++) {
				String typ = in.readUTF();
				if(typ.equals("Umbuchung")) {
					Umbuchung umbuchung = new Umbuchung();
					umbuchung.laden(in, db);
					Register[] register = (Register[]) umbuchung.getKategorie();
					register[1] = this;
					register[0].einsortierenBuchung(umbuchung);
					if(register[0] != register[1])
						register[1].einsortierenBuchung(umbuchung);
					umbuchung.setKategorie(register);
					if(DEBUG)
						System.out.println("Umbuchung einsortiert: "+register[0]+"/"+register[1]);
				}
				else if(typ.equals("StandardBuchung2")) {
					StandardBuchung standardBuchung = new StandardBuchung();
					standardBuchung.laden(in, db);
					einsortierenBuchung(standardBuchung);
					db.buchungMerken(standardBuchung);
				}
				else if(typ.equals("StandardBuchung")) {
					Datum datum = new Datum();
					datum.laden(in);
					String text = in.readUTF();
					int anz = in.readInt();
					if(anz == 1) {
						Kategorie kategorie = db.findeOderErzeugeKategorie(in.readUTF());
						Euro betrag = new Euro();
						betrag.laden(in);
						StandardBuchung standardBuchung = new StandardBuchung(datum, text, kategorie, betrag); 
						einsortierenBuchung(standardBuchung);
						db.buchungMerken(standardBuchung);
					}
					else {
						SplitBuchung buchung = new SplitBuchung(datum, text);
						for(int j=0; j<anz; j++) {
							String kategorie = in.readUTF();
							Euro betrag = new Euro();
							betrag.laden(in);
							buchung.add(db.findeOderErzeugeKategorie(kategorie), betrag);
						}
						einsortierenBuchung(buchung);
            db.buchungMerken(buchung);
					}
				}
				else if(typ.equals("SplitBuchung")) {
					SplitBuchung splitBuchung = new SplitBuchung();
					splitBuchung.laden(in, db);
					einsortierenBuchung(splitBuchung);
          db.buchungMerken(splitBuchung);
				}
				else if(!typ.equals("Dummy"))
					throw new IOException("Register.laden: Unbekannter Buchungstyp: "+typ);
				// "Dummy"-Buchung wird ignoriert!
			}
		}

		void speichern(DataOutputStream out)
			throws IOException {
			out.writeUTF(name);
			out.writeInt(buchungen.size());
			for(int i=0; i<buchungen.size(); i++) {
				AbstractBuchung buchung = getBuchung(i);
				if(buchung.getClass() == Umbuchung.class) {
					Register[] register = (Register[]) ((Umbuchung)buchung).getKategorie();
					if((register[0] == this) && (register[0] != register[1])) {
  					out.writeUTF("Dummy");
  					// Umbuchungen werden nur vom Zielregister gespeichert.
 						// "Dummy"-Buchung wird geschrieben damit die Anzahl stimmt!
  					// Selbstbuchungen dennoch schreiben!
 						if(DEBUG)
 						  System.out.println("Register "+name+": Dummy-Buchung gespeichert.");
					}
  				else
  					buchung.speichern(out);
				}
				else
  				buchung.speichern(out);
			}
		}

    String[][] csvExport() {
      int anzahl = getAnzahlBuchungen();
      String[][] text = new String[anzahl][3];
      for(int i=0; i<anzahl; i++) {
        AbstractBuchung buchung = getBuchung(i);
        text[i][0] = ""+buchung.getDatum();
        text[i][1] = buchung.getText();
        text[i][2] = ""+buchung.getKategorie();
        text[i][3] = ""+buchung.getWert();
      }
      return text;
    }

}
